from pony.orm import Database, Required, Optional, Json, LongStr, Set
from datetime import datetime

db = Database()
    
class Target(db.Entity):
    #   ----- Attributes -----

    target_name = Required(str, unique=True)  # Required: Cannot be None

    #   ----- Relations -----

    exposures = Set('Exposure')  # One target contains a set of exposures
    processed_exposure = Set('Processed_Exposure')

# Exposure table class
class Exposure(db.Entity):
    #   ----- Attributes -----

    observation_time = Required(datetime, unique=True)
    obs_id = Required(int, size=32, unsigned=True)
    insMode = Required(str)
    datacube_header = Optional(Json)
    raw_exposure_header = Optional(Json)
    raw_exposure_data = Optional(Json)
    raw_exposure_filename = Optional(str, unique=True)
    prm_filename = Optional(str, unique=True)
    pampelmuse_params = Optional(Json)
    sources = Optional(Json)
    psf_table = Optional(Json)
    pampelmuse_catalog = Optional(Json)
    raman_image_header = Optional(Json)

    #   ----- Sky parameters -----
    sky_condition_start_time = Optional(float)
    sky_condition_start = Optional(LongStr)
    sky_comment_start = Optional(LongStr)
    sky_condition_end_time = Optional(float)
    sky_condition_end = Optional(LongStr)
    sky_comment_end = Optional(LongStr)

    #   ----- Relations -----

    target = Required('Target')  # One exposure belongs to a target
    processed_exposure = Optional('Processed_Exposure')

class Processed_Exposure(db.Entity):
    observation_time = Required(datetime, unique=True)
    obs_id = Required(int, size=32, unsigned=True)
    insMode = Required(str)
    raw_filename = Optional(str, unique=True)
    # ------- Maybe some of those can be grouped in one field like atm params ------
    ngs_flux = Optional(float) # Not sure what data type it should be
    ttfree = Optional(bool)
    degraded = Optional(bool)
    glf = Optional(float)
    seeing = Optional(float)
    seeing_los = Optional(float)
    airMass = Optional(float)
    tau0 = Optional(float)
    # --------------------------------------------------------------
    num_sources = Optional(int, unsigned=True)
    sgs_data = Optional(Json) # sgs_data extension
    ag_data = Optional(Json) # ag_data extension
    sparta_cn2 = Optional(Json) # sparta_cn2 extension
    sparta_atm =Optional(Json) # sparta_atm extension
    psf_params = Optional(Json)
    sparta_iq_data = Optional(Json)
    sparta_iq_los_500nm = Optional(float)
    sparta_iq_los_500nm_nogain = Optional(float)
    
    # Relations
    
    target = Required('Target')  # One exposure belongs to a target
    exposure = Required('Exposure')

def create_conn(db, provider, host, user, passwd, database):
    #connect to the database
    db.bind(provider=provider, host=host, user=user, passwd=passwd, db=database)
    db.generate_mapping(create_tables = True)


from pony.orm import *
from AOMUSE.db.Database import Exposure, Processed_Exposure, db, create_conn
import pandas as pd
import json
from astropy import units as u
from astropy.coordinates import SkyCoord
from AOMUSE.utils.preprocess import get_sparta_iq_los, fwhm_from500nm_to_l
import numpy as np

class muse_db():

    def __init__(self, db, provider, host, user, passwd, database):
        create_conn(db, provider=provider, host=host, user=user, passwd=passwd, db=database)

    def process(self):
        pes = select(pe for pe in Processed_Exposure)
        pes.delete()
        exposures = select(e for e in Exposure)

        #To count how many exposures have  a bad fit in them
        allStars = 0
        for expo in exposures:
            dfExpos = {}
            #json return dictionaries
            #data contains the headers of different type of files (raw and reduced)
            try:
                primary = json.loads(expo.datacube_header)
            except Exception as e:
                print("Exposure skipped")
                print("datacube_header =", expo.datacube_header, "(Probably empty string)")
                single_exposures_skipped += 1
                continue
                
            try:
                data = json.loads(expo.raw_exposure_data)
            except Exception as e:
                print("Exposure skipped")
                print("raw_exposure_data =", expo.raw_exposure_data, "(Probably empty string)")
                raw_exposures_skipped += 1
                continue
            
            
            #stars contain the info of the sources used on the PampelMUSE fit
            try:
                stars = json.loads(expo.sources)
            except Exception as e:
                print("Exposure skipped")
                print("sources =", expo.sources, "(Probably empty string)")
                stars_skipped += 1
                continue
            
            #psfparams has the PampelMUSE polyfit parameters
            psfParams = json.loads(expo.pampelmuse_params)
            dfPars = pd.DataFrame(psfParams)
            #check if all is good
            if(not(dfPars.isnull().values.any())):
                #filename
                dfExpos['observation_time'] = expo.observation_time
                dfExpos['obs_id'] = expo.obs_id
                dfExpos['insMode'] = expo.insMode
                dfExpos['target'] = expo.target
                dfExpos['exposure'] = expo
                dfExpos['raw_filename'] = expo.raw_exposure_filename
                dfStars = {}
                for star in stars.items():
                    dfStar = pd.DataFrame(star[1])
                    if(not(dfStar.isnull().values.any())):
                        dfStars[star[0]] = dfStar 
                if(dfStars == {}):
                    allStars += 1
                    continue
                #ngs_flux is not available for WFM
                dfExpos['ngs_flux'] = -1
                
                # we check whether we are in tt free mode
                # HIERARCH ESO AOS TT LOOP ST = T / Tip-tilt loop status
                if not bool(primary['ESO AOS TT LOOP ST']):
                    dfExpos['ttfree'] = 1
                else:
                    dfExpos['ttfree'] = 0
                try:
                    all_lasers_on = not bool(primary['ESO LGS1 LASR1 SWSIM']) and \
                                    not bool(primary['ESO LGS2 LASR2 SWSIM']) and \
                                    not bool(primary['ESO LGS3 LASR3 SWSIM']) and \
                                    not bool(primary['ESO LGS4 LASR4 SWSIM'])
                    if not all_lasers_on:
                        dfExpos['degraded'] = 1
                    else:
                        dfExpos['degraded'] = 0
                except:
                    dfExpos['degraded'] = -1
                #ground layer fraction
                mass_dimm_glf = primary['ESO OCS SGS ASM GL900 AVG']
                dfExpos["glf"] = mass_dimm_glf

                # Equatorial coordinates of the observation
                ra = primary['RA']*u.degree
                dec = primary['DEC']*u.degree
                coords_J2000 = SkyCoord(ra,dec)
                
                air_mass = (primary['ESO TEL AIRM START'] + primary['ESO TEL AIRM END'])/2.0
                dimm_seeing = (primary['ESO TEL AMBI FWHM START'] + primary['ESO TEL AMBI FWHM END'])/2.0
                dfExpos['seeing'] = dimm_seeing
                dfExpos['airMass'] = air_mass
                #correct seeing for airmass
                dfExpos['seeing_los'] = dimm_seeing*air_mass**(3./5.)
            
                #coherence time
                dfExpos['tau0'] = primary['ESO TEL AMBI TAU0']
                
                #fwhm for different wavelenght ranges, i call them u,v and i just cause they go from blue to infrared
                
                #number of sources in the PPMUSE input catalogue
                dfExpos['num_sources'] = len(stars)
                
                # Slow Guidance System, SGS, data
                dfExpos['sgs_data'] = data['SGS_DATA']
                
                # Telescope Guide probe data
                dfExpos['ag_data'] = data['AG_DATA']
            
                if 'SPARTA_CN2_DATA' in data and \
                'SPARTA_ATM_DATA' in data and \
                    not bool(primary['ESO AOS TT LOOP ST']):
                    sparta_iq_data = {}
                    
                    dfExpos['sparta_cn2'] = data['SPARTA_CN2_DATA']
                    dfExpos['sparta_atm'] = data['SPARTA_ATM_DATA']

                    # we add in quadrature the contributions of GALACSI and MUSE
                    # For MUSE I use the IQ vs wavelength in the Performance Report 
                    # which was better than the TLRs:
                    # 0.37" for 480-600 nm
                    # 0.29" for 600-800 nm
                    # 0.29" for 800-930 nm
                    #
                    try:
                        sparta_iq_los_500nm = get_sparta_iq_los(coords_J2000, data['SPARTA_ATM_DATA'], mass_dimm_glf)
                        dfExpos['sparta_iq_los_500nm'] = sparta_iq_los_500nm
                    except Exception as e:
                        print(f"{e}")
                        dfExpos['sparta_iq_los_500nm'] = 0.0
                    skip_i = []
                    psf_params = {}
                    
                    wavelength = psfParams["wavelength"]
                    windows = range(0, len(wavelength), int(len(wavelength)/10))
                    key = f"wavelength_means"
                    wavelength_means = []
                    for i in range(len(windows)-1):
                        if(wavelength[windows[i]] > 576 and wavelength[windows[i]] < 605 or wavelength[windows[i+1]] > 576 and wavelength[windows[i+1]] < 605):
                            skip_i.append(i)
                            continue
                        mean = np.mean([wavelength[windows[i]], wavelength[windows[i+1]]])
                        wavelength_means.append(mean)
                    psf_params["wavelength"] = wavelength_means
                    sparta_iq_data[key] = wavelength_means
                    for par_name in psfParams.keys():
                        if(par_name == "wavelength"):
                            continue
                        param = psfParams[par_name]

                        windows = range(0, len(param), int(len(param)/10))
                        key = f"{par_name}_means"
                        means = []
                        for i in range(len(windows)-1):
                            if(i in skip_i):
                                continue
                            mean = np.mean([param[windows[i]], param[windows[i+1]]])
                            means.append(mean)
                        psf_params[key] = means
                    dfExpos['psf_params'] = psf_params
                    ra = primary['RA']*u.degree
                    dec = primary['DEC']*u.degree
                    coords_J2000 = SkyCoord(ra,dec)
                    mass_dimm_glf = primary['ESO OCS SGS ASM GL900 AVG']
                    sparta_iq_los_500nm = get_sparta_iq_los(coords_J2000, data['SPARTA_ATM_DATA'], mass_dimm_glf)
                    sparta_iq_los_500nm_nogain = get_sparta_iq_los(coords_J2000, data['SPARTA_ATM_DATA'], mass_dimm_glf, True)
                    dfExpos['sparta_iq_los_500nm'] = sparta_iq_los_500nm
                    dfExpos['sparta_iq_los_500nm_nogain'] = sparta_iq_los_500nm_nogain
                    l = []
                    l_nogain = []
                    for wavelength in wavelength_means:
                        l.append(fwhm_from500nm_to_l(sparta_iq_los_500nm, wavelength, True))
                        l_nogain.append(fwhm_from500nm_to_l(sparta_iq_los_500nm_nogain, wavelength, True))
                    sparta_iq_data["sparta_iq_l"] = l
                    sparta_iq_data["sparta_iq_l_nogain"] = l_nogain
                    dfExpos['sparta_iq_data'] = sparta_iq_data
                keys = []
                for key, value in dfExpos.items():
                    if(value is None):
                        keys.append(key)
                for key in keys:
                    del dfExpos[key]
                Processed_Exposure(**dfExpos)

    def get_exposures(self):
        es = select(e for e in Exposure)
        data = []
        columns = [
            "observation_time", "target", "obs_id", "insMode", 
            "raw_filename", "prm_filename", 
            "exposure", "processed_exposure"
        ]
        for e in es:
            data.append([e.observation_time, e.target.target_name, e.obs_id, e.insMode, 
                        e.raw_exposure_filename, e.prm_filename, 
                        e, e.processed_exposure
                        ])
        df = pd.DataFrame(data, columns=columns)
        return df

    def get_proocessed_exposures(self):
        pes = select(pe for pe in Processed_Exposure)
        data = []
        columns = [
            "observation_time", "target", "obs_id", "insMode", 
            "raw_filename", "exposure", "processed_exposure"
        ]
        for pe in pes:
            data.append([pe.observation_time, pe.target.target_name, pe.obs_id, pe.insMode, 
                        pe.raw_filename, pe.exposure, pe
                        ])
        df = pd.DataFrame(data, columns=columns)
        return df

    def get_processed_data(self):
        pes = select(pe for pe in Processed_Exposure)
        data = []
        columns = [
            "observation_time", "obs_id", "insMode", "raw_filename",
            "ngs_flux", "ttfree", "degraded", "glf", "seeing",
            "seeing_los", "airMass", "tau0", "num_sources",
            "sparta_iq_los_500nm", "sparta_iq_los_500nm_nogain"
        ]
        for pe in pes:
            data.append([pe.observation_time, pe.obs_id, pe.insMode, pe.raw_filename,
                        pe.ngs_flux, pe.ttfree, pe.degraded, pe.glf, pe.seeing,
                        pe.seeing_los, pe.airMass, pe.tau0, pe.num_sources,
                        pe.sparta_iq_los_500nm, pe.sparta_iq_los_500nm_nogain
                        ])
        df = pd.DataFrame(data, columns=columns)
        return df

    def get_processed_tables(self):
        pes = select(pe for pe in Processed_Exposure)
        tables = []
        columns = ["observation_time", "target", "obs_id", "insMode", 
            "sgs_data", "ag_data", "sparta_cn2", "sparta_atm", "psf_params", "sparta_iq_data"]

        for pe in pes:
            sgs_data = pd.DataFrame(pe.sgs_data)
            ag_data = pd.DataFrame(pe.ag_data)
            sparta_cn2 = pd.DataFrame(pe.sparta_cn2)
            sparta_atm = pd.DataFrame(pe.sparta_atm)
            psf_params = pd.DataFrame(pe.psf_params)
            sparta_iq_data = pd.DataFrame(pe.sparta_iq_data)
            tables.append([pe.observation_time, pe.target.target_name, pe.obs_id, pe.insMode, 
                        sgs_data, ag_data, sparta_cn2, sparta_atm, psf_params, sparta_iq_data])
        df = pd.DataFrame(tables, columns=columns)
        return df

    


import numpy as np
from astropy.time import Time
from astropy.coordinates import AltAz
from AOMUSE.constants import location

# Procedure to get the fwhm at an arbitrary wavelength including
# aberrations introduced by MUSE.
def fwhm_from500nm_to_l(fwhm, l, muse_aberr=False):
    """
    function to correct the seeing by
    wavelenght
    """
    
    # Note that for now we only add 0.3" in quadrature
    if muse_aberr:
        return np.sqrt((fwhm*(500.0/l)**0.2)**2 + 0.35**2)
    else:
        return fwhm*(500.0/l)**0.2
    
def get_sparta_iq_los(coords_J2000, sparta, mass_dimm_glf, no_ao_gain=False):
    """
    function to compute IQ from the LOS LGS telemetry
    """
    try:
        # Martinez, Kolb, Sarazin, Tokovinin (2010, The Messenger 141, 5) 
        # originally provided by Tokovinin (2002, PASP 114, 1156) but 
        # corrected by Kolb (ESO Technical Report #12)
        #
        # Note that we also have two sections that can be used to determine
        # the exact date and time of the telemetry data. Together with the
        # RA, Dec and coordinates of Paranal we can get the exact airmass
        # of eac measurement. The SPARTA QC0 script uses all this info.
    
        times_atmos = Time(sparta['Sec'], format='unix')
        times_atmos.format='isot'
        airmass_atmos = []
        for obstime in times_atmos:
            current_coords_altaz = coords_J2000.transform_to(\
                                AltAz(obstime=obstime,location=location))
            z_atmos = current_coords_altaz.zen
            airmass_atmos = np.append(airmass_atmos, 1/np.cos(z_atmos))

        airmass = np.nanmedian(airmass_atmos)
    
        L0 = (np.mean(sparta["LGS1_L0"]) + np.mean(sparta["LGS2_L0"]) + \
              np.mean(sparta["LGS3_L0"]) + np.mean(sparta["LGS4_L0"]))/4.0
        R0 = (np.mean(sparta["LGS1_R0"]) + np.mean(sparta["LGS2_R0"]) + \
              np.mean(sparta["LGS3_R0"]) + np.mean(sparta["LGS4_R0"]))/4.0
        sparta_glf = (np.nanmean(sparta["LGS1_TUR_GND"]) + np.nanmean(sparta["LGS2_TUR_GND"]) + \
              np.nanmean(sparta["LGS3_TUR_GND"]) + np.nanmean(sparta["LGS4_TUR_GND"]))/4.0
    
        Fkolb = 1.0/(300.0*8.0/L0) - 1.0
    
        # Note: The model sets FWHMatm=0 if the argument of the square root becomes 
        # negative [1+FKolb⋅2.183⋅(r0/L0)0.356]<0 , which happens when the Fried 
        # parameter r0 reaches its threshold of rt=L0⋅[1/(2.183⋅FKolb)]1/0.356. For the
        # VLT and L0=46m , this corresponds to rt=5.4m.
        #
        if 1+Fkolb*2.183*(R0/L0)**0.356 <= 0.0:
            Fcorr = 0.0
        else:
            Fcorr = np.sqrt(1+Fkolb*2.183*(R0/L0)**0.356)

        sparta_seeing = 1.028*np.rad2deg(500.0e-9/R0)*3600.0
        #sparta_seeing = (np.mean(sparta["LGS1_SEEING"]) + np.mean(sparta["LGS2_SEEING"]) + \
           #np.mean(sparta["LGS3_SEEING"]) + np.mean(sparta["LGS4_SEEING"]))/4.0

        if no_ao_gain:
            expected_ao_gain = 1.0
        else:
            #expected_ao_gain = polynomial_ao_gain(sparta_glf)
            expected_ao_gain = polynomial_ao_gain(mass_dimm_glf)
    
        # what is the LHSi_FWHM_GAIN? we do not use it in Milli's QC0 script
        # sparta_ao_gain = (np.mean(sparta['LGS1_FWHM_GAIN']) + np.mean(sparta['LGS2_FWHM_GAIN']) + \
        #       np.mean(sparta['LGS3_FWHM_GAIN']) + np.mean(sparta['LGS4_FWHM_GAIN']))/4.0

        # print('sparta_seeing: ', sparta_seeing, 'Fcorr: ', Fcorr, 'sparta_ao_gain: ', sparta_ao_gain)
        #
        # Note that airmass is not used. If observation has a large range of airmasses
        # the simle averaging that we are using is not good. The muse_sparta_qc0 script 
        # uses the time stamp and ra, dec to get airmass of each sampled point.
        # This should be reviewed.
        #
        sparta_iq_los =  sparta_seeing*Fcorr/expected_ao_gain
    except Exception as e:
        print(f"{e}")
        sparta_iq_los =  0.0
    
    return sparta_iq_los

# This is the AO gain implemented in the QC0 script and
# in OT. The ETC uses the mean value.
#
def polynomial_ao_gain(x):
    return 1.09 + 0.42*x -1.98*x**2 + 2.95*x**3


from pony.orm import *
from AOMUSE.db.Database import Exposure, Processed_Exposure, db, create_conn
import pandas as pd
import json
from astropy import units as u
from astropy.coordinates import SkyCoord
from AOMUSE.utils.preprocess import get_sparta_iq_los, fwhm_from500nm_to_l
import numpy as np

pes = select(pe for pe in Processed_Exposure)
pes.delete()
exposures = select(e for e in Exposure)

#To count how many exposures have  a bad fit in them
parsNaN = 0
allStars = 0
for expo in exposures:
    dfExpos = {}
    #json return dictionaries
    #data contains the headers of different type of files (raw and reduced)
    try:
        primary = json.loads(expo.datacube_header)
    except Exception as e:
        #print(f"{e}")
        print("Exposure skipped")
        print("datacube_header =", expo.datacube_header, "(Probably empty string)")
        single_exposures_skipped += 1
        continue
        
    try:
        data = json.loads(expo.raw_exposure_data)
    except Exception as e:
        #print(f"{e}")
        print("Exposure skipped")
        print("raw_exposure_data =", expo.raw_exposure_data, "(Probably empty string)")
        raw_exposures_skipped += 1
        continue
    
    
    #stars contain the info of the sources used on the PampelMUSE fit
    try:
        stars = json.loads(expo.sources)
    except Exception as e:
        #print(f"{e}")
        print("Exposure skipped")
        print("sources =", expo.sources, "(Probably empty string)")
        stars_skipped += 1
        continue
    
    #psfparams has the PampelMUSE polyfit parameters
    psfParams = json.loads(expo.pampelmuse_params)
    dfPars = pd.DataFrame(psfParams)
    #check if all is good
    if(not(dfPars.isnull().values.any())):
        #filename
        dfExpos['observation_time'] = expo.observation_time
        dfExpos['obs_id'] = expo.obs_id
        dfExpos['insMode'] = expo.insMode
        dfExpos['target'] = expo.target
        dfExpos['exposure'] = expo
        dfExpos['raw_filename'] = expo.raw_exposure_filename
        #print(len(stars))
        dfStars = {}
        for star in stars.items():
            dfStar = pd.DataFrame(star[1])
            if(not(dfStar.isnull().values.any())):
                dfStars[star[0]] = dfStar 
        if(dfStars == {}):
            #print(f"{expo.singleFile} bad sources")
            allStars += 1
            continue
        pars = star[1].keys()
        #ngs_flux is not available for WFM
        dfExpos['ngs_flux'] = -1
        
        # we check whether we are in tt free mode
        # HIERARCH ESO AOS TT LOOP ST = T / Tip-tilt loop status
        if not bool(primary['ESO AOS TT LOOP ST']):
            dfExpos['ttfree'] = 1
        else:
            dfExpos['ttfree'] = 0
            
        # We check weather we are in degraded mode
        # HIERARCH ESO AOS LGS1 DET GAIN = 100 / Detector gain of the LGSi
        # HIERARCH ESO LGS1 LASR1 OPMODE = 3 / Operational mode : Normal, Maintenance,Serv
        # HIERARCH ESO LGS1 LASR1 REPUMPER ST = T / True when the re-pumper is active.
        # HIERARCH ESO LGS1 LASR1 SWSIM = F / If T, function is software simulated.
        # HIERARCH ESO LGS1 LASR1 POWER = 22.321 / Laser power.
        # HIERARCH ESO LGS1 SHUT1 ST = T / Shutter open.
        #dfExpos['degraded'] = 0
        try:
            all_lasers_on = not bool(primary['ESO LGS1 LASR1 SWSIM']) and \
                            not bool(primary['ESO LGS2 LASR2 SWSIM']) and \
                            not bool(primary['ESO LGS3 LASR3 SWSIM']) and \
                            not bool(primary['ESO LGS4 LASR4 SWSIM'])
            if not all_lasers_on:
                dfExpos['degraded'] = 1
            else:
                dfExpos['degraded'] = 0
        except:
            dfExpos['degraded'] = -1
        #ground layer fraction
        mass_dimm_glf = primary['ESO OCS SGS ASM GL900 AVG']
        dfExpos["glf"] = mass_dimm_glf

        # Equatorial coordinates of the observation
        ra = primary['RA']*u.degree
        dec = primary['DEC']*u.degree
        coords_J2000 = SkyCoord(ra,dec)
        
        air_mass = (primary['ESO TEL AIRM START'] + primary['ESO TEL AIRM END'])/2.0
        dimm_seeing = (primary['ESO TEL AMBI FWHM START'] + primary['ESO TEL AMBI FWHM END'])/2.0
        dfExpos['seeing'] = dimm_seeing
        dfExpos['airMass'] = air_mass
        #correct seeing for airmass
        dfExpos['seeing_los'] = dimm_seeing*air_mass**(3./5.)
    
        #coherence time
        dfExpos['tau0'] = primary['ESO TEL AMBI TAU0']
        
        #fwhm for different wavelenght ranges, i call them u,v and i just cause they go from blue to infrared
        
        #number of sources in the PPMUSE input catalogue
        dfExpos['num_sources'] = len(stars)
        
        # Slow Guidance System, SGS, data
        dfExpos['sgs_data'] = data['SGS_DATA']
        
        # Telescope Guide probe data
        dfExpos['ag_data'] = data['AG_DATA']
    
        if 'SPARTA_CN2_DATA' in data and \
           'SPARTA_ATM_DATA' in data and \
            not bool(primary['ESO AOS TT LOOP ST']):
            sparta_iq_data = {}
            
            dfExpos['sparta_cn2'] = data['SPARTA_CN2_DATA']
            dfExpos['sparta_atm'] = data['SPARTA_ATM_DATA']

            # we add in quadrature the contributions of GALACSI and MUSE
            # For MUSE I use the IQ vs wavelength in the Performance Report 
            # which was better than the TLRs:
            # 0.37" for 480-600 nm
            # 0.29" for 600-800 nm
            # 0.29" for 800-930 nm
            #
            try:
                sparta_iq_los_500nm = get_sparta_iq_los(coords_J2000, data['SPARTA_ATM_DATA'], mass_dimm_glf)
                dfExpos['sparta_iq_los_500nm'] = sparta_iq_los_500nm
            except Exception as e:
                print(f"{e}")
                dfExpos['sparta_iq_los_500nm'] = 0.0
            skip_i = []
            psf_params = {}
            
            wavelength = psfParams["wavelength"]
            windows = range(0, len(wavelength), int(len(wavelength)/10))
            key = f"wavelength_means"
            wavelength_means = []
            for i in range(len(windows)-1):
                if(wavelength[windows[i]] > 576 and wavelength[windows[i]] < 605 or wavelength[windows[i+1]] > 576 and wavelength[windows[i+1]] < 605):
                    skip_i.append(i)
                    continue
                mean = np.mean([wavelength[windows[i]], wavelength[windows[i+1]]])
                wavelength_means.append(mean)
            psf_params["wavelength"] = wavelength_means
            sparta_iq_data[key] = wavelength_means
            for par_name in psfParams.keys():
                if(par_name == "wavelength"):
                    continue
                param = psfParams[par_name]

                windows = range(0, len(param), int(len(param)/10))
                key = f"{par_name}_means"
                means = []
                for i in range(len(windows)-1):
                    if(i in skip_i):
                        continue
                    mean = np.mean([param[windows[i]], param[windows[i+1]]])
                    means.append(mean)
                psf_params[key] = means
            dfExpos['psf_params'] = psf_params
            ra = primary['RA']*u.degree
            dec = primary['DEC']*u.degree
            coords_J2000 = SkyCoord(ra,dec)
            mass_dimm_glf = primary['ESO OCS SGS ASM GL900 AVG']
            sparta_iq_los_500nm = get_sparta_iq_los(coords_J2000, data['SPARTA_ATM_DATA'], mass_dimm_glf)
            sparta_iq_los_500nm_nogain = get_sparta_iq_los(coords_J2000, data['SPARTA_ATM_DATA'], mass_dimm_glf, True)
            dfExpos['sparta_iq_los_500nm'] = sparta_iq_los_500nm
            dfExpos['sparta_iq_los_500nm_nogain'] = sparta_iq_los_500nm_nogain
            l = []
            l_nogain = []
            for wavelength in wavelength_means:
                l.append(fwhm_from500nm_to_l(sparta_iq_los_500nm, wavelength, True))
                l_nogain.append(fwhm_from500nm_to_l(sparta_iq_los_500nm_nogain, wavelength, True))
            sparta_iq_data["sparta_iq_l"] = l
            sparta_iq_data["sparta_iq_l_nogain"] = l_nogain
            dfExpos['sparta_iq_data'] = sparta_iq_data
        keys = []
        for key, value in dfExpos.items():
            if(value is None):
                keys.append(key)
        for key in keys:
            del dfExpos[key]
        Processed_Exposure(**dfExpos)


from astropy import units as u
from astropy import coordinates

# Definition of the UT4 coordinates
latitude =  -24.6270*u.degree
longitude = -70.4040*u.degree  
altitude = 2648.0*u.meter 
location = coordinates.EarthLocation(lon=longitude,lat=latitude,height=altitude)